//  Diffusion Client Library for iOS, tvOS and OS X / macOS
//
//  Copyright (c) 2015, 2018 Push Technology Ltd., All Rights Reserved.
//
//  Use is subject to license terms.
//
//  NOTICE: All information contained herein is, and remains the
//  property of Push Technology. The intellectual and technical
//  concepts contained herein are proprietary to Push Technology and
//  may be covered by U.S. and Foreign Patents, patents in process, and
//  are protected by trade secret or copyright law.

@import Foundation;
#import <Diffusion/PTDiffusionFeature.h>
#import <Diffusion/PTDiffusionTopicUnsubscriptionReason.h>

@class PTDiffusionStream;
@class PTDiffusionTopicDetails;
@class PTDiffusionTopicDetailsLevel;
@class PTDiffusionTopicSelector;
@class PTDiffusionTopicStream;
@class PTDiffusionValueStream;

@protocol PTDiffusionTopicStreamDelegate;
@protocol PTDiffusionFetchStreamDelegate;

NS_ASSUME_NONNULL_BEGIN

/**
 @brief The Topics feature provides a client session with the capability to receive streamed
 topic updates and/or fetch the state of topics.

 The Topics feature for a session can be obtained from the session's
 `topics` property.

 @see PTDiffusionSession

 @since 5.6
 */
@interface PTDiffusionTopicsFeature : PTDiffusionFeature

/**
 Fetch the current state of topics.
 
 @param expression The @ref md_topic_selectors "topic selector" expression to be
 evaluated by the server.
 
 @param delegate Receiver of messages relating to this fetch request.
 The delegate will be sent messages asynchronously on the main dispatch queue.
 
 @return A new stream instance configured to notify the delegate of events.
 This does not need to be retained by the caller, but should be if later use of
 the removeStream: method is required.
 
 @note This method will return `nil` if the session has been closed.

 @exception NSInvalidArgumentException raised if any supplied arguments are `nil`.

 @since 5.6
 */
-(nullable PTDiffusionStream *)fetchWithTopicSelectorExpression:(NSString *)expression
                                                       delegate:(id<PTDiffusionFetchStreamDelegate>)delegate;

/**
 Get the details of the given topic.
 
 @param topicPath The full path of the topic to be queried.
 
 @param level The level of detail required in the response.
 
 @param completionHandler Block to be called asynchronously on success or failure.
 If the operation was successful, the `error` argument passed to the block will be `nil`.
 When successful the `topicDetails` argument passed to the block will be `nil` if the topic
 is unknown.
 The completion handler will be called asynchronously on the main dispatch queue.

 @exception NSInvalidArgumentException raised if any supplied arguments are `nil`.
 
 @deprecated Since 6.0. The use of PTDiffusionTopicDetails to describe topics is
 being phased out. PTDiffusionTopicSpecification is now being used in
 preference and is supplied on subscription with value streams. This method will
 be removed in a future release.

 @since 5.7
 */
-(void)getTopicDetailsWithTopicPath:(NSString *)topicPath
                              level:(PTDiffusionTopicDetailsLevel *)level
                  completionHandler:(void (^)(PTDiffusionTopicDetails * _Nullable topicDetails, NSError * _Nullable error))completionHandler
__deprecated_msg("Will be removed in a future release.");

/**
 Request subscription to topics.
 
 @param expression The @ref md_topic_selectors "topic selector" expression to be
 evaluated by the server.

 @param completionHandler Block to be called asynchronously on success or failure.
 If the operation was successful, the `error` argument passed to the block will be `nil`.
 The completion handler will be called asynchronously on the main dispatch queue.

 @exception NSInvalidArgumentException Raised if any supplied arguments are `nil`.

 @since 5.6
 */
-(void)subscribeWithTopicSelectorExpression:(NSString *)expression
                          completionHandler:(void (^)(NSError * _Nullable error))completionHandler;

/**
 Request unsubscription from topics.

 @param expression The @ref md_topic_selectors "topic selector" expression to be
 evaluated by the server.

 @param completionHandler Block to be called asynchronously on success or failure.
 If the operation was successful, the `error` argument passed to the block will be `nil`.
 The completion handler will be called asynchronously on the main dispatch queue.

 @exception NSInvalidArgumentException Raised if any supplied arguments are `nil`.

 @since 5.6
 */
-(void)unsubscribeFromTopicSelectorExpression:(NSString *)expression
                            completionHandler:(void (^)(NSError * _Nullable error))completionHandler;

/**
 Add a topic stream to receive topic updates for topics that match the given topic selector.

 When a topic update is received from the server, it will be passed to all topic streams that have
 been added with matching selectors.
 If there is more than one match, the order in which the topic streams are called is not defined.
 If there is no matching topic stream, the fallback topic stream(s) will be called instead.
 
 @param selector The @ref md_topic_selectors "topic selector" to be evaluated locally.

 @param delegate Receiver of messages on matching topic stream events.
 Used by the stream to notify of events but not retained.
 The delegate will be sent messages asynchronously on the main dispatch queue.

 @return A new stream instance configured to notify the delegate of events.
 This does not need to be retained by the caller, but should be if later use of
 the removeStream: method is required.

 @exception NSInvalidArgumentException Raised if any supplied arguments are `nil`.

 @note Since the addition of value streams for version 5.9, this method now
 returns instances of PTDiffusionTopicStream, being a new subclass of
 PTDiffusionStream.
 
 @deprecated Since 6.0. Topic streams are only necessary to support the
 deprecated stateless and record topic types. They will be removed along with
 these topic types in a future release.

 @since 5.6
 */
-(PTDiffusionTopicStream *)addTopicStreamWithSelector:(PTDiffusionTopicSelector *)selector
                                             delegate:(id<PTDiffusionTopicStreamDelegate>)delegate
__deprecated_msg("Will be removed in a future release.");

/**
 A convenience method wrapping a call to `[self addTopicStreamWithSelector:delegate]`
 
 @param expression The @ref md_topic_selectors "topic selector" expression to be
 evaluated locally.
 
 @param delegate Receiver of messages on matching topic stream events.
 Used by the stream to notify of events but not retained.
 The delegate will be sent messages asynchronously on the main dispatch queue.

 @exception NSInvalidArgumentException Raised if any supplied arguments are `nil`.

 @return A new stream instance configured to notify the delegate of events.
 This does not need to be retained by the caller, but should be if later use of
 the removeStream: method is required.

 @deprecated Since 6.0. Topic streams are only necessary to support the
 deprecated stateless and record topic types. They will be removed along with
 these topic types in a future release.

 @since 5.9
 */
-(PTDiffusionTopicStream *)addTopicStreamWithSelectorExpression:(NSString *)expression
                                                       delegate:(id<PTDiffusionTopicStreamDelegate>)delegate
__deprecated_msg("Will be removed in a future release.");

/**
 When a topic update is received from the server, it will be passed on to all topic streams that have been
 added with matching selectors. If no topic stream is registered with a matching selector, the
 fallback topic streams that have been registered using this method will be called instead.

 Zero, one, or many fallback topic streams can be set. 
 If there is no fallback topic stream, updates that match no other will be discarded.
 
 @param delegate Receiver of messages on topic stream events not handled by an explicitly registered topic selector.
 Used by the stream to notify of events but not retained.
 The delegate will be sent messages asynchronously on the main dispatch queue.

 @return A new topic stream instance configured to notify the delegate of events.
 This does not need to be retained by the caller, but should be if later use of
 the removeStream: method is required.

 @exception NSInvalidArgumentException Raised if the supplied `delegate` argument is `nil`.

 @note Since the addition of value streams for version 5.9, this method now
 returns instances of PTDiffusionTopicStream, being a new subclass of
 PTDiffusionStream.
 
 @deprecated Since 6.0. Topic streams are only necessary to support the
 deprecated stateless and record topic types. They will be removed along with
 these topic types in a future release.

 @since 5.6
 */
-(PTDiffusionTopicStream *)addFallbackTopicStreamWithDelegate:(id<PTDiffusionTopicStreamDelegate>)delegate
__deprecated_msg("Will be removed in a future release.");

/**
 When a matching update is received from the server for a topic, it will be
 passed on to all value streams that have been added with matching selectors.
 If no value stream is registered with a matching selector, the fallback value
 streams that have been registered using this method will be called instead.

 Zero, one, or many fallback value streams can be set.
 If there is no fallback value stream, updates that match no other will be
 discarded.

 @param stream Handler of value stream events not handled by an explicitly
 registered topic selector.
 The delegate associated with this stream will be sent messages asynchronously
 on the main dispatch queue.

 @exception NSInvalidArgumentException Raised if the supplied `stream` argument
 is `nil`.

 @since 5.9
 */
-(void)addFallbackStream:(PTDiffusionValueStream *)stream;

/**
 Add a value stream to receive topic updates for topics that match the given
 topic selector.

 When a value update is received from the server, it will be passed to all
 streams that have been added with matching selectors.
 If there is more than one match, the order in which the streams are called is
 not defined.
 If there is no matching stream, the fallback stream(s) will be called instead.

 @param stream Handler of value stream events against this selector.
 The delegate associated with this stream will be sent messages asynchronously
 on the main dispatch queue.

 @param selector The @ref md_topic_selectors "topic selector" to be evaluated locally.

 @exception NSInvalidArgumentException Raised if any supplied arguments are `nil`.

 @since 5.9
 */
-(void)addStream:(PTDiffusionValueStream *)stream
    withSelector:(PTDiffusionTopicSelector *)selector;

/**
 Add a value stream to receive topic updates for topics that match the given
 topic selector expression.

 This convenience method calls addStream:withSelector: having constructed a
 PTDiffusionTopicSelector instance using the given expression.

 @param stream Handler of value stream events against this selector.
 The delegate associated with this stream will be sent messages asynchronously
 on the main dispatch queue.

 @param expression The @ref md_topic_selectors "topic selector" expression to be
 evaluated locally.

 @exception NSInvalidArgumentException Raised if any supplied arguments are `nil`.

 @since 6.0
 */
-(void)          addStream:(PTDiffusionValueStream *)stream
    withSelectorExpression:(NSString *)expression;

/**
 Remove a topic or value stream that was previously added using a topic selector
 or as a fallback.

 @param stream The stream to remove.

 @return `YES` if the stream was removed.

 @note If the supplied `stream` argument is `nil` or the stream has already been
 removed from the feature then calls to this method do nothing and return `NO`.

 @since 5.9
 */
-(BOOL)removeStream:(nullable PTDiffusionStream *)stream;

@end

NS_ASSUME_NONNULL_END
